@@ -4,7 +4,7 @@
  * (pharvey@codebydesign.com).
  *
  * Modified and extended by Nick Gorham
- * (nick@easysoft.com).
+ * (nick@lurcher.org).
  *
  * Any bugs or problems should be considered the fault of Nick and not
  * Peter.
@@ -175,16 +175,16 @@ SQLRETURN SQLColumnsW( SQLHSTMT statement_handle,
 
     thread_protect( SQL_HANDLE_STMT, statement );
 
-    if (( name_length1 < 0 && name_length1 != SQL_NTS ) ||
-            ( name_length2 < 0 && name_length2 != SQL_NTS ) ||
-            ( name_length3 < 0 && name_length3 != SQL_NTS ) ||
-            ( name_length4 < 0 && name_length4 != SQL_NTS ))
+    if (( catalog_name && name_length1 < 0 && name_length1 != SQL_NTS ) ||
+            ( schema_name && name_length2 < 0 && name_length2 != SQL_NTS ) ||
+            ( table_name && name_length3 < 0 && name_length3 != SQL_NTS ) ||
+            ( column_name && name_length4 < 0 && name_length4 != SQL_NTS ))
     {
         __post_internal_error( &statement -> error,
                 ERROR_HY090, NULL,
                 statement -> connection -> environment -> requested_version );
 
-        return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR );
+        return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
     }
 
     /*
@@ -210,11 +210,14 @@ SQLRETURN SQLColumnsW( SQLHSTMT statement_handle,
                 ERROR_24000, NULL,
                 statement -> connection -> environment -> requested_version );
 
-        return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR );
+        return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
     }
     else if ( statement -> state == STATE_S8 ||
             statement -> state == STATE_S9 ||
-            statement -> state == STATE_S10 )
+            statement -> state == STATE_S10 ||
+            statement -> state == STATE_S13 ||
+            statement -> state == STATE_S14 ||
+            statement -> state == STATE_S15 )
     {
         dm_log_write( __FILE__,
                 __LINE__,
@@ -226,7 +229,7 @@ SQLRETURN SQLColumnsW( SQLHSTMT statement_handle,
                 ERROR_HY010, NULL,
                 statement -> connection -> environment -> requested_version );
 
-        return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR );
+        return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
     }
 
     if ( statement -> state == STATE_S11 ||
@@ -238,7 +241,7 @@ SQLRETURN SQLColumnsW( SQLHSTMT statement_handle,
                     ERROR_HY010, NULL,
                     statement -> connection -> environment -> requested_version );
 
-            return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
         }
     }
 
@@ -261,7 +264,7 @@ SQLRETURN SQLColumnsW( SQLHSTMT statement_handle,
                 ERROR_IM001, NULL,
                     statement -> connection -> environment -> requested_version );
 
-            return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
         }
 
         ret = SQLCOLUMNSW( statement -> connection ,
@@ -278,6 +281,7 @@ SQLRETURN SQLColumnsW( SQLHSTMT statement_handle,
     else
     {
         SQLCHAR *as1, *as2, *as3, *as4;
+        int clen;
 
         if ( !CHECK_SQLCOLUMNS( statement -> connection ))
         {
@@ -291,13 +295,17 @@ SQLRETURN SQLColumnsW( SQLHSTMT statement_handle,
                 ERROR_IM001, NULL,
                     statement -> connection -> environment -> requested_version );
 
-            return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
         }
 
-        as1 = (SQLCHAR*) unicode_to_ansi_alloc( catalog_name, name_length1, statement -> connection );
-        as2 = (SQLCHAR*) unicode_to_ansi_alloc( schema_name, name_length2, statement -> connection );
-        as3 = (SQLCHAR*) unicode_to_ansi_alloc( table_name, name_length3, statement -> connection );
-        as4 = (SQLCHAR*) unicode_to_ansi_alloc( column_name, name_length4, statement -> connection );
+        as1 = (SQLCHAR*) unicode_to_ansi_alloc( catalog_name, name_length1, statement -> connection, &clen );
+        name_length1 = clen;
+        as2 = (SQLCHAR*) unicode_to_ansi_alloc( schema_name, name_length2, statement -> connection, &clen );
+        name_length2 = clen;
+        as3 = (SQLCHAR*) unicode_to_ansi_alloc( table_name, name_length3, statement -> connection, &clen );
+        name_length3 = clen;
+        as4 = (SQLCHAR*) unicode_to_ansi_alloc( column_name, name_length4, statement -> connection, &clen );
+        name_length4 = clen;
 
         ret = SQLCOLUMNS( statement -> connection ,
                 statement -> driver_stmt,
@@ -330,7 +338,7 @@ SQLRETURN SQLColumnsW( SQLHSTMT statement_handle,
 
         if ( ret == SQL_SUCCESS_WITH_INFO )
         {
-            function_return_ex( SQL_HANDLE_STMT, statement, ret, TRUE );
+            function_return_ex( IGNORE_THREAD, statement, ret, TRUE, DEFER_R1 );
         }
         
         SQLNUMRESULTCOLS( statement -> connection,
@@ -366,5 +374,5 @@ SQLRETURN SQLColumnsW( SQLHSTMT statement_handle,
                 statement -> msg );
     }
 
-    return function_return( SQL_HANDLE_STMT, statement, ret );
+    return function_return( SQL_HANDLE_STMT, statement, ret, DEFER_R1 );
 }
