@@ -4,7 +4,7 @@
  * (pharvey@codebydesign.com).
  *
  * Modified and extended by Nick Gorham
- * (nick@easysoft.com).
+ * (nick@lurcher.org).
  *
  * Any bugs or problems should be considered the fault of Nick and not
  * Peter.
@@ -224,7 +224,10 @@ SQLRETURN SQLGetDescRecW( SQLHDESC descriptor_handle,
         __check_stmt_from_desc( descriptor, STATE_S9 ) ||
         __check_stmt_from_desc( descriptor, STATE_S10 ) ||
         __check_stmt_from_desc( descriptor, STATE_S11 ) ||
-        __check_stmt_from_desc( descriptor, STATE_S12 )) {
+        __check_stmt_from_desc( descriptor, STATE_S12 ) ||
+        __check_stmt_from_desc( descriptor, STATE_S13 ) ||
+        __check_stmt_from_desc( descriptor, STATE_S14 ) ||
+        __check_stmt_from_desc( descriptor, STATE_S15 )) {
 
         dm_log_write( __FILE__, 
                 __LINE__, 
@@ -236,7 +239,7 @@ SQLRETURN SQLGetDescRecW( SQLHDESC descriptor_handle,
                 ERROR_HY010, NULL,
                 descriptor -> connection -> environment -> requested_version );
 
-        return function_return( SQL_HANDLE_DESC, descriptor, SQL_ERROR );
+        return function_return_nodrv( SQL_HANDLE_DESC, descriptor, SQL_ERROR );
     }
 
     if( __check_stmt_from_desc_ird( descriptor, STATE_S1 )) {
@@ -251,7 +254,7 @@ SQLRETURN SQLGetDescRecW( SQLHDESC descriptor_handle,
                 ERROR_HY007, NULL,
                 descriptor -> connection -> environment -> requested_version );
 
-        return function_return( SQL_HANDLE_DESC, descriptor, SQL_ERROR );
+        return function_return_nodrv( SQL_HANDLE_DESC, descriptor, SQL_ERROR );
     }
 
 
@@ -270,7 +273,7 @@ SQLRETURN SQLGetDescRecW( SQLHDESC descriptor_handle,
                     ERROR_IM001, NULL,
                     descriptor -> connection -> environment -> requested_version );
 
-            return function_return( SQL_HANDLE_DESC, descriptor, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_DESC, descriptor, SQL_ERROR );
         }
 
         ret = SQLGETDESCRECW( descriptor -> connection,
@@ -302,7 +305,7 @@ SQLRETURN SQLGetDescRecW( SQLHDESC descriptor_handle,
                     ERROR_IM001, NULL,
                     descriptor -> connection -> environment -> requested_version );
 
-            return function_return( SQL_HANDLE_DESC, descriptor, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_DESC, descriptor, SQL_ERROR );
         }
 
         if ( name && buffer_length > 0 )
@@ -325,7 +328,7 @@ SQLRETURN SQLGetDescRecW( SQLHDESC descriptor_handle,
 
         if ( SQL_SUCCEEDED( ret ) && name && as1 )
         {
-            ansi_to_unicode_copy( name, (char*) as1, SQL_NTS, descriptor -> connection );
+            ansi_to_unicode_copy( name, (char*) as1, SQL_NTS, descriptor -> connection, NULL );
         }
         if ( as1 )
         {
@@ -349,7 +352,7 @@ SQLRETURN SQLGetDescRecW( SQLHDESC descriptor_handle,
 \n\t\t\tScale = %s\
 \n\t\t\tNullable = %s",
                     __get_return_status( ret, s8 ),
-                    __sdata_as_string( s1, SQL_CHAR, 
+                    __sdata_as_string( s1, SQL_WCHAR, 
                         string_length, name ),
                     __sptr_as_string( s2, type ),
                     __sptr_as_string( s3, sub_type ),
@@ -365,5 +368,5 @@ SQLRETURN SQLGetDescRecW( SQLHDESC descriptor_handle,
                 descriptor -> msg );
     }
 
-    return function_return( SQL_HANDLE_DESC, descriptor, ret );
+    return function_return( SQL_HANDLE_DESC, descriptor, ret, DEFER_R3 );
 }
