@@ -4,7 +4,7 @@
  * (pharvey@codebydesign.com).
  *
  * Modified and extended by Nick Gorham
- * (nick@easysoft.com).
+ * (nick@lurcher.org).
  *
  * Any bugs or problems should be considered the fault of Nick and not
  * Peter.
@@ -177,10 +177,10 @@ SQLRETURN SQLProcedureColumnsW(
 
     thread_protect( SQL_HANDLE_STMT, statement );
 
-    if (( cb_catalog_name < 0 && cb_catalog_name != SQL_NTS ) ||
-            ( cb_schema_name < 0 && cb_schema_name != SQL_NTS ) ||
-            ( cb_proc_name < 0 && cb_proc_name != SQL_NTS ) ||
-            ( cb_column_name < 0 && cb_column_name != SQL_NTS ))
+    if (( sz_catalog_name && cb_catalog_name < 0 && cb_catalog_name != SQL_NTS ) ||
+            ( sz_schema_name && cb_schema_name < 0 && cb_schema_name != SQL_NTS ) ||
+            ( sz_proc_name && cb_proc_name < 0 && cb_proc_name != SQL_NTS ) ||
+            ( sz_column_name && cb_column_name < 0 && cb_column_name != SQL_NTS ))
     {
         dm_log_write( __FILE__, 
                 __LINE__, 
@@ -192,7 +192,7 @@ SQLRETURN SQLProcedureColumnsW(
                 ERROR_HY090, NULL,
                 statement -> connection -> environment -> requested_version );
 
-        return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR );
+        return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
     }
 
     /*
@@ -218,11 +218,14 @@ SQLRETURN SQLProcedureColumnsW(
                 ERROR_24000, NULL,
                 statement -> connection -> environment -> requested_version );
 
-        return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR );
+        return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
     }
     else if ( statement -> state == STATE_S8 ||
             statement -> state == STATE_S9 ||
-            statement -> state == STATE_S10 )
+            statement -> state == STATE_S10 ||
+            statement -> state == STATE_S13 ||
+            statement -> state == STATE_S14 ||
+            statement -> state == STATE_S15 )
     {
         dm_log_write( __FILE__, 
                 __LINE__, 
@@ -234,7 +237,7 @@ SQLRETURN SQLProcedureColumnsW(
                 ERROR_HY010, NULL,
                 statement -> connection -> environment -> requested_version );
 
-        return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR );
+        return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
     }
 
     if ( statement -> state == STATE_S11 ||
@@ -252,7 +255,7 @@ SQLRETURN SQLProcedureColumnsW(
                     ERROR_HY010, NULL,
                     statement -> connection -> environment -> requested_version );
 
-            return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
         }
     }
 
@@ -275,7 +278,7 @@ SQLRETURN SQLProcedureColumnsW(
                     ERROR_IM001, NULL,
                     statement -> connection -> environment -> requested_version );
 
-            return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
         }
 
         ret = SQLPROCEDURECOLUMNSW( statement -> connection ,
@@ -292,6 +295,7 @@ SQLRETURN SQLProcedureColumnsW(
     else
     {
         SQLCHAR *as1, *as2, *as3, *as4;
+        int clen;
 
         if ( !CHECK_SQLPROCEDURECOLUMNS( statement -> connection ))
         {
@@ -305,13 +309,17 @@ SQLRETURN SQLProcedureColumnsW(
                     ERROR_IM001, NULL,
                     statement -> connection -> environment -> requested_version );
 
-            return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
         }
 
-        as1 = (SQLCHAR*) unicode_to_ansi_alloc( sz_catalog_name, cb_catalog_name, statement -> connection );
-        as2 = (SQLCHAR*) unicode_to_ansi_alloc( sz_schema_name, cb_schema_name, statement -> connection );
-        as3 = (SQLCHAR*) unicode_to_ansi_alloc( sz_proc_name, cb_proc_name, statement -> connection );
-        as4 = (SQLCHAR*) unicode_to_ansi_alloc( sz_column_name, cb_column_name, statement -> connection );
+        as1 = (SQLCHAR*) unicode_to_ansi_alloc( sz_catalog_name, cb_catalog_name, statement -> connection, &clen );
+        cb_catalog_name = clen;
+        as2 = (SQLCHAR*) unicode_to_ansi_alloc( sz_schema_name, cb_schema_name, statement -> connection, &clen );
+        cb_schema_name = clen;
+        as3 = (SQLCHAR*) unicode_to_ansi_alloc( sz_proc_name, cb_proc_name, statement -> connection, &clen );
+        cb_proc_name = clen;
+        as4 = (SQLCHAR*) unicode_to_ansi_alloc( sz_column_name, cb_column_name, statement -> connection, &clen );
+        cb_column_name = clen;
 
         ret = SQLPROCEDURECOLUMNS( statement -> connection ,
                 statement -> driver_stmt,
@@ -360,5 +368,5 @@ SQLRETURN SQLProcedureColumnsW(
                 statement -> msg );
     }
 
-    return function_return( SQL_HANDLE_STMT, statement, ret );
+    return function_return( SQL_HANDLE_STMT, statement, ret, DEFER_R1 );
 }
