@@ -13,12 +13,14 @@
  *
  */
 
+#include <config.h>
 #include "environ.h"
 #include "connection.h"
 #include "socket.h"
 #include "statement.h"
 #include "qresult.h"
 #include "lobj.h"
+#include "misc.h"
 #include "dlg_specific.h"
 #include <stdio.h>
 #include <string.h>
@@ -1095,6 +1097,15 @@ char cmdbuffer[MAX_MESSAGE_LEN+1];	/* QR_set_command() dups this string so dont
 
 				while( ! clear) {
 					id = SOCK_get_char(sock);
+					if ((SOCK_get_errcode(sock) != 0) || (id == EOF)) {
+						CC_set_error(self, CONNECTION_NO_RESPONSE, "No response from the backend");
+						if (res) {
+							QR_Destructor(res);
+						}
+						mylog("send_query: id=%d error=%s \n", id, "No response from the backend");
+						CC_set_no_trans(self);
+						return NULL;
+					}
 					switch(id) {
 					case 'I':
 						(void) SOCK_get_char(sock);
@@ -1186,7 +1197,7 @@ char cmdbuffer[MAX_MESSAGE_LEN+1];	/* QR_set_command() dups this string so dont
 			mylog("send_query: 'E' - %s\n", msgbuffer);
 			qlog("ERROR from backend during send_query: '%s'\n", msgbuffer);
 
-			/* We should report that an error occured. Zoltan */
+			/* We should report that an error occurred. Zoltan */
             if (res)
                 QR_Destructor(res);
 			res = QR_Constructor();
